import pretty_output, argparse, traceback

from karamel.command import Command

from kooki.config import read_config_file, get_kooki_recipe_manager, get_kooki_dir_recipes, get_kooki_jar_manager, get_kooki_dir_jars
from kooki.rule_parser import parse_document_rules
from kooki.exception import KookiException
from kooki.process import process_document

from karamel.packages import install_packages, freeze_packages

__command__ = 'bake'
__description__ = 'Bake a kooki'


class BakeCommand(Command):

    def __init__(self):
        super(BakeCommand, self).__init__(__command__, __description__)
        self.add_argument('documents', nargs='*')
        self.add_argument('--config-file', default='kooki.yaml')
        self.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS, help='Show this help message and exit.')
        self.add_argument('-d', '--debug', help='Show information to help debug the bake processing', action='store_true')
        self.add_argument('--no-color', help='The output has no color.', action='store_true')
        self.add_argument('--no-output', help='There is no output.', action='store_true')

    def callback(self, args):
        try:
            pretty_output.set_output_policy(not args.no_output)
            pretty_output.set_color_policy(not args.no_color)
            pretty_output.set_debug_policy(args.debug)

            self.real_call(args)

        except KookiException as e:
            pretty_output.error_step('Errors')
            pretty_output.error(e)

        except Exception as e:
            pretty_output.error_step('Errors')
            pretty_output.error(traceback.format_exc()[:-1])

    def real_call(self, args):
        config = read_config_file(args.config_file)
        document_rules = parse_document_rules(config)

        if args.documents == []:
            documents = document_rules
        else:
            documents = {}
            for document_name in args.documents:
                if document_name in document_rules:
                    documents[document_name] = document_rules[document_name]
                else:
                    raise Exception('Bad document')

        for name, document in documents.items():
            pretty_output.title_1(name)
            self.execute_unfreeze(document)
            self.execute_bake(document)
            pretty_output.title_2()

        if len(documents) == 0:
            pretty_output.info('nothing to do')

    def on_package_downloading(self, package_name):
        print('Downloading \'{}\''.format(package_name))

    def on_package_installing(self, package_name):
        print('Installing \'{}\''.format(package_name))

    def on_package_install_success(self, package_name):
        print('Successfully installed \'{}\''.format(package_name))

    def on_package_already_installed(self, package):
        def callback(package_name, package_path):
            print('{} \'{}\' already installed in \'{}\'.'.format(package, package_name, package_path))
        return callback

    def on_package_not_found(self, package):
        def callback(package_name):
            raise KookiException('{} not found \'{}\'.'.format(package, package_name))
        return callback

    def on_package_bad_version_provided(self, package):
        def callback(self, package_name, version):
            raise KookiException('Could not find the version \'{1}\' for {2} \'{0}\'.'.format(package_name, version, package))
        return callback

    def on_package_could_not_be_download(self, package):
        def callback(self, package_name):
            message = \
    '''{} \'{}\' could not be download.
    Without this package the document cannot be generated.
    Are you connected to the Internet ?'''.format(package_name)
            raise KookiException(message.format(package, package_name))
        return callback

    def execute_unfreeze(self, document):
        pretty_output.title_2('unfreeze')

        pretty_output.title_3('jars')
        package_manager_url = get_kooki_jar_manager()
        package_install_dir = get_kooki_dir_jars()
        install_packages(package_manager_url,
                         package_install_dir,
                         document.jars,
                         self.on_package_downloading,
                         self.on_package_installing,
                         self.on_package_install_success,
                         self.on_package_already_installed('Jar'),
                         self.on_package_not_found('Jar'),
                         self.on_package_bad_version_provided('jar'),
                         self.on_package_could_not_be_download('Jar'))

        pretty_output.title_3('recipe')
        package_manager_url = get_kooki_recipe_manager()
        package_install_dir = get_kooki_dir_recipes()
        install_packages(package_manager_url,
                         package_install_dir,
                         [document.recipe],
                         self.on_package_downloading,
                         self.on_package_installing,
                         self.on_package_install_success,
                         self.on_package_already_installed('Recipe'),
                         self.on_package_not_found('Recipe'),
                         self.on_package_bad_version_provided('recipe'),
                         self.on_package_could_not_be_download('Recipe'))

    def execute_bake(self, document):
        pretty_output.title_2('bake')
        process_document(document)



import textwrap, yaml
from collections import OrderedDict
from kooki.exception import KookiException
from kooki.tools import write_file

invalid_field_message = textwrap.dedent('''\
    Invalid config file
    The '{1}' field of '{0}' is not in a valid format.
    It should be a {2} and it is a {3}.''')


class Rule:

    reserved = ['name', 'template', 'recipe', 'jars', 'metadata', 'content']

    def __init__(self, document_name='', config={}):
        self.document_name = document_name
        self.name = ''
        self.template = ''
        self.recipe = ''
        self.jars = []
        self.metadata = []
        self.content = []
        self.set_config(config)

    def set_config(self, config):
        for key, value in config.items():
            if key == 'name':
                if not isinstance(value, str):
                    raise KookiException(invalid_field_message.format(self.document_name, key, str, type(value)))
                self.name = value
            elif key == 'template':
                if not isinstance(value, str):
                    raise KookiException(invalid_field_message.format(self.document_name, key, str, type(value)))
                self.template = value
            elif key == 'recipe':
                if not isinstance(value, str):
                    raise KookiException(invalid_field_message.format(self.document_name, key, str, type(value)))
                self.recipe = value
            elif key == 'jars':
                if not isinstance(value, list):
                    raise KookiException(invalid_field_message.format(self.document_name, key, list, type(value)))
                self.jars += value
            elif key == 'metadata':
                if not isinstance(value, list):
                    raise KookiException(invalid_field_message.format(self.document_name, key, list, type(value)))
                self.metadata += value
            elif key == 'content':
                if not isinstance(value, list):
                    raise KookiException(invalid_field_message.format(self.document_name, key, list, type(value)))
                self.content += value

    def copy(self):
        rule = Rule()
        rule.name = self.name
        rule.template = self.template
        rule.recipe = self.recipe
        rule.jars = list(self.jars)
        rule.metadata = list(self.metadata)
        rule.content = list(self.content)
        return rule

    def export(self):

        from yaml import Loader, Dumper

        def dict_representer(dumper, data):
            return dumper.represent_dict(data.items())

        Dumper.add_representer(OrderedDict, dict_representer)

        config = OrderedDict()
        config[self.name] = OrderedDict()
        config[self.name]['name'] = self.name
        config[self.name]['template'] = self.template
        config[self.name]['recipe'] = self.recipe
        config[self.name]['jars'] = self.jars
        config[self.name]['metadata'] = self.metadata
        config[self.name]['content'] = self.content

        config_file = yaml.dump(config, Dumper=Dumper, default_flow_style=False)
        return config_file[:-1]

    def set_name(self, name, default):
        if name:
            if not isinstance(name, str):
                raise KookiException(invalid_field_message.format(self.name, 'name', str, type(name)))
            self.name = name
        else:
            self.name = default

    def set_recipe(self, recipe, default):
        if recipe:
            if not isinstance(recipe, str):
                raise KookiException(invalid_field_message.format(self.name, 'recipe', str, type(recipe)))
            self.recipe = recipe
        else:
            self.recipe = default

    def set_template(self, template, default):
        if template:
            if not isinstance(template, str):
                raise KookiException(invalid_field_message.format(self.name, 'template', str, type(template)))
            self.template = template
        else:
            self.template = default

    def set_jars(self, jars, default):
        if jars:
            if not isinstance(jars, list):
                raise KookiException(invalid_field_message.format(self.name, 'jars', list, type(jars)))
            for jar in jars:
                self.jars.append(jar)
        else:
            self.jars = default

    def set_metadata(self, metadata_list, default):
        if metadata_list:
            if not isinstance(metadata_list, list):
                raise KookiException(invalid_field_message.format(self.name, 'metadata', list, type(metadata_list)))
            for metadata in metadata_list:
                self.metadata.append(metadata)
        else:
            write_file('metadata.yaml', default)
            self.metadata.append('metadata.yaml')

    def set_content(self, content_list, default):
        if content_list:
            if not isinstance(content_list, list):
                raise KookiException(invalid_field_message.format(self.name, 'content', list, type(content_list)))
            for content in content_list:
                self.content.append(content)
        else:
            write_file('content.md', default)
            self.content.append('content.md')



import pretty_output, os

from collections import OrderedDict

from kooki.jars import search_file, search_recipe, search_jar
from kooki.extension import load_extensions, caller
from kooki.exception import MissingRecipe, MissingTemplate, MissingMetadata, MissingContent, InvalidRecipe
from kooki.common import Metadata

output = []

def output_search_start():
    global output
    output = []


def output_search(path, fullpath):
    if fullpath:
        output.append({'name': path, 'status': '[found]', 'path': fullpath})
    else:
        output.append({'name': path, 'status': ('[missing]', 'red'), 'path': ''})


def output_search_finish():
    pretty_output.infos(output, [('name', 'blue'), ('status', 'green'), ('path', 'cyan')])


def process_document(document_origin):
    document = document_origin.copy()
    recipe = document.recipe
    jars = document.jars

    # jars
    full_path_jars = []
    for jar in document.jars:
        jar_full_path = search_jar(jar)
        full_path_jars.append(jar_full_path)
    document.jars = full_path_jars

    # extensions
    pretty_output.title_3('extensions')
    output_search_start()
    extensions_raw = load_extensions(document.jars, document.recipe)
    extensions = Metadata()
    for extension_name, extension_path in extensions_raw.items():

        splitted = extension_name.split('.')
        current = extensions
        for s in splitted[:-1]:
            if not s in current:
                current[s] = Metadata()
            current = current[s]
        current[splitted[-1]] = caller(extension_path)

        output_search(extension_name, extension_path)
    output_search_finish()

    # template
    pretty_output.title_3('template')
    output_search_start()
    file_full_path = search_file(jars, recipe, document.template)
    output_search(document.template, file_full_path)
    if file_full_path:
        with open(file_full_path,'r') as f:
            file_read = f.read()
        document.template = file_read
        output_search_finish()
    else:
        output_search_finish()
        raise MissingTemplate(jars, recipe, document.template)

    # metadata
    pretty_output.title_3('metadata')
    metadata_full_path = {}
    output_search_start()
    for metadata in document.metadata:
        file_full_path = search_file(jars, recipe, metadata)
        output_search(metadata, file_full_path)
        if file_full_path:
            with open(file_full_path,'r') as f:
                file_read = f.read()
            metadata_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingMetadata(jars, recipe, metadata)
    document.metadata = metadata_full_path
    output_search_finish()

    # content
    pretty_output.title_3('content')
    content_full_path = OrderedDict()
    output_search_start()
    for content in document.content:
        file_full_path = search_file(jars, recipe, content)
        output_search(content, file_full_path)
        if file_full_path:
            with open(file_full_path,'r') as f:
                file_read = f.read()
            content_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingContent(jars, recipe, content)
    document.content = content_full_path
    output_search_finish()

    # recipe
    pretty_output.title_3('recipe')
    recipe_full_path = search_recipe(recipe)

    if not recipe_full_path or not os.path.isdir(recipe_full_path):
        raise MissingRecipe(recipe)

    try:
        import imp
        fp, pathname, description = imp.find_module('recipe', [recipe_full_path])
        recipe_module = imp.load_module('recipe', fp, pathname, description)
        recipe_module.recipe(document, extensions)

    except AttributeError as e:
        raise InvalidRecipe(recipe)

    except ImportError:
        print('cant import')
        raise InvalidRecipe(recipe)
